#include <Wire.h>
#include <FS.h>
#include <SD.h>
#include <SPI.h>
#include "image.h"
#include "AXP192.h"
#include "ILI9342C.h"
#include "LoadJpeg.h"
#include "TouchPanel.h"

#define GAMESTAT_SEL 0
#define GAMESTAT_PLAY 1
#define GAMESTAT_COMP 2

#define DEBUG false

ILI9342C ili9342c ;
LoadJPEG image ;
TouchPanel touch ;

int imageNo = 1 ;
int BTN = 0 ;
int POS = 0 ;
int imgNo = 0 ;
int gameStat = GAMESTAT_SEL ;
int panel[16] ;
int bx = 3 ;
int by = 3 ;
int bp = 15 ;
unsigned int seed = 0;

void setup() {

#if DEBUG
  Serial.begin(9600) ;
  delay(2000) ;
  Serial.println( ) ;
  Serial.println( ) ;
#endif

  // --- I/F INIT
  SPI.begin();  //SPIを初期化、SCK、MOSI、SSの各ピンの動作は出力、SCK、MOSIはLOW、SSはHIGH 
  Wire1.begin(21, 22);

  // --- INIT POWER
  AXP192 apx ;  // コンストラクタが動けばお仕事完了
  apx.setBreath(0xcc) ; // 撮影用に画面を暗めにしておく

  // --- ILI9342C
  ili9342c.init() ;

  // --- IMAGE
  image.initBuff( ) ;
  ReadImageFile( ) ;
  image.overLayImage( ) ;

  // --- DISP IMAGE
  ili9342c.dispBitMap(0 , 0 , 320, 240, image.imageBuff) ;
}

void ReadImageFile( ) {
  char fileName[16] ;
  // --- IMAGE
  image.initBuff( ) ;
  if (gameStat != GAMESTAT_COMP) {
    imgNo ++ ;
  }
  if (imgNo != 0) {
    sprintf(fileName,"/image%d.jpg",imgNo) ;
    if (!image.LoadImage(fileName)) {
      imgNo = 0 ;
      image.initBuff( ) ;
      image.overLayImage( ) ;
    }
  } else {
    image.overLayImage( ) ;
  }
}

void loop() {
  seed ++ ;
  delay(100) ;

  if (touch.readPoint( )) {
    // --- 何処かが押されてる
#if DEBUG
    Serial.print("BTN=") ;
    Serial.println(touch.BTN) ;
#endif
    if (touch.BTN == 1) {
      ReadImageFile( ) ;
      // --- DISP IMAGE
      ili9342c.dispBitMap(0 , 0 , 320, 240, image.imageBuff) ;
      gameStat = GAMESTAT_SEL ;
    } else
    if (touch.BTN == 3) {
      image.makePanelImage( ) ;
      gameStat = GAMESTAT_PLAY ;
      shafflePanel( ) ;
      for (int i=0;i<16;i++) {
        dispPanle(i) ;
      }
      delay(1000) ;
    }
    if (touch.BTN == 0 && gameStat == GAMESTAT_PLAY) {
      if (touch.POSSTAT == STAT_TOUCH) {
#if DEBUG
        Serial.print("Touch:X=") ;
        Serial.print(touch.POSX) ;
        Serial.print("  Y=") ;
        Serial.print(touch.POSY) ;
        Serial.print("  STAT=") ;
        Serial.println(touch.POSSTAT) ;
        Serial.print("BlankPanel:X=") ;
        Serial.print(bx) ;
        Serial.print("  Y=") ;
        Serial.println(by) ;
#endif
        if ((touch.POSY != by || touch.POSX != bx) && (touch.POSY == by || touch.POSX == bx)) {
          // パネル移動
          if (touch.POSX == bx) {
            // 縦移動
            if (touch.POSY > by) {
              // ■□□□
              for (int y=by;y<touch.POSY;y++) {
                int p1 = y * 4 + bx ;
                int p2 = (y+1) * 4 + bx ;
                // SWAP 
                int w = panel[p1] ;
                panel[p1] = panel[p2] ;
                panel[p2] = w ;
                dispPanle(p1) ;
              }
              by = touch.POSY ;
              bp = by * 4 + bx ;
              dispPanle(bp) ;
            } else {
              // □□□■
              for (int y=by;y>touch.POSY;y--) {
                int p1 = y * 4 + bx ;
                int p2 = (y-1) * 4 + bx ;
                // SWAP 
                int w = panel[p1] ;
                panel[p1] = panel[p2] ;
                panel[p2] = w ;
                dispPanle(p1) ;
              }
              by = touch.POSY ;
              bp = by * 4 + bx ;
              dispPanle(bp) ;
            }
          } else {
            // 横移動
            if (touch.POSX > bx) {
              // ■□□□
              for (int x=bx;x<touch.POSX;x++) {
                int p1 = by * 4 + x ;
                int p2 = by * 4 + x + 1 ;
                // SWAP 
                int w = panel[p1] ;
                panel[p1] = panel[p2] ;
                panel[p2] = w ;
                dispPanle(p1) ;
              }
              bx = touch.POSX ;
              bp = by * 4 + bx ;
              dispPanle(bp) ;
            } else {
              // □□□■
              for (int x=bx;x>touch.POSX;x--) {
                int p1 = by * 4 + x ;
                int p2 = by * 4 + x - 1 ;
                // SWAP 
                int w = panel[p1] ;
                panel[p1] = panel[p2] ;
                panel[p2] = w ;
                dispPanle(p1) ;
              }
              bx = touch.POSX ;
              bp = by * 4 + bx ;
              dispPanle(bp) ;
            }
          }
        }
        if (checkClear( )) {
          // --- CLEAR ---
          gameStat = GAMESTAT_COMP ;
          ReadImageFile( ) ;
          image.overLayImage2( ) ;
          ili9342c.dispBitMap(0 , 0 , 320, 240, image.imageBuff) ;
          delay(2000) ;
          ReadImageFile( ) ;
          ili9342c.dispBitMap(0 , 0 , 320, 240, image.imageBuff) ;
          gameStat = GAMESTAT_SEL ;
        }
      }
    }
    touch.resetPos( ) ;
  }
}

// ===== GAME本体 =====
void dispPanle(int pn) {
  int xs = (pn % 4) * 80 ;
  int ys = pn / 4 * 60 ;
  ili9342c.dispBitMap(xs , ys , 80, 60, image.panelPtr[panel[pn]]) ;
}

void shafflePanel( ) {
  bx = 3 ;
  by = 3 ;
  bp = 15 ;
  randomSeed(seed) ;
  for (int i=0;i<16;i++) {
    panel[i] = i;
  }
  int suffleNum = 0 ;
  while(suffleNum < 10) {
    int p1 = random(15);
    int p2 = random(15);
    if (p1 != p2) {
      int t = panel[p1] ;
      panel[p1] = panel[p2] ;
      panel[p2] = t ;
      suffleNum ++ ;
    }
  }
}

bool checkClear( ) {
  bool result = true ;
  for (int i=0;i<16;i++) {
    if (panel[i] != i) {
      result = false ;
      break ;
    }
  }
  return result ;
}
