// ILI9342 SPI Graphics TEST
#include <SPI.h>
#include "ILI9342C.h"

#define TFT_CLK 18
#define TFT_MISO 38
#define TFT_MOSI 23
#define TFT_DC 15
#define TFT_CS 5

//#define TFT_RST 9   // AXP_IO4
//#define TFT_BL 7    // AXP_DC3
//#define TFT_PWR 7   // AXP_LDO2

#define ILI9342_RDMODE 0x0A     ///< Read Display Power Mode
#define ILI9342_RDMADCTL 0x0B   ///< Read Display MADCTL
#define ILI9342_RDPIXFMT 0x0C   ///< Read Display Pixel Format
#define ILI9342_RDIMGFMT 0x0D   ///< Read Display Image Format
#define ILI9342_RDSELFDIAG 0x0F ///< Read Display Self-Diagnostic Result

#define ILI9342_SOFTRESET 0x01              // SOFTWARE RESET
#define ILI9342_SLEEP_OUT 0x11              // SLEEP OUT
#define ILI9342_DISPLAY_OFF 0x28            // DISPLAY OFF
#define ILI9342_DISPLAY_ON 0x29             // DISPLAY ON
#define ILI9342_COLUMN_ADDRESS_SET 0x2a     // COL SET
#define ILI9342_PAGE_ADDRESS_SET 0x2b       // ROW SET
#define ILI9342_MEMORY_WRITE 0x2c           // MEMORY WRITE
#define ILI9342_MEMORY_ACCESS_CONTROL 0x36  // 表示向き

#define ILI9342_NORON 0x13    ///< Normal Display Mode ON
#define ILI9342_PTLON 0x12    ///< Partial Mode ON
#define ILI9342_INVOFF 0x20   ///< Display Inversion OFF
#define ILI9342_INVON 0x21    ///< Display Inversion ON

#define ILI9342_COLMOD 0x3a                 // 表示モード

#define ILI9342_COLOR16 0x55                // 1Dot/16Bitモード
#define ILI9342_COLOR18 0x66                // 1Dot/18Bitモード
#define ILI9342_MAC 0xe8   // 0xe8 / 0x48  // 表示方向　-　横(0xe8)/縦(0x48)
#define ILI9342_WIDTH 320  // 320  / 240
#define ILI9342_HEIGHT 240 // 240  / 320

int tftWidth ;
int tftHeight ;
uint8_t tftReadBuff[16] ;
uint8_t SPIBuf[320*2] ;   // SPI転送用バッファ

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
SPISettings settings = SPISettings(16000000, MSBFIRST, SPI_MODE0);

void ILI9342C::init() {
  // ---- SPI
//  SPI.beginTransaction(settings);
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(SPI_MODE0);

  // -----  ILI9342C INITIAL -----
  pinMode(TFT_CS, OUTPUT);
  pinMode(TFT_DC, OUTPUT);
//  pinMode(TFT_RST, OUTPUT); // M5Stack CORE2だとRESETはAXP192に接続しているので面倒くさい

  lcd_init( ) ;
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
// TFTにコマンドを送信
void ILI9342C::tftSendCommand(uint8_t command) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// TFTにコマンド+1バイトデータを送信
void ILI9342C::tftSendCommand1(uint8_t command, uint8_t data1) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer(data1);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// TFTにコマンド+2バイトデータを送信
void ILI9342C::tftSendCommand2(uint8_t command, uint8_t data1, uint8_t data2) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer(data1);
  SPI.transfer(data2);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// TFTにコマンド+3バイトデータを送信
void ILI9342C::tftSendCommand3(uint8_t command, uint8_t data1, uint8_t data2, uint8_t data3) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer(data1);
  SPI.transfer(data2);
  SPI.transfer(data3);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// TFTにコマンド+4バイトデータを送信
void ILI9342C::tftSendCommand4(uint8_t command, uint8_t data1, uint8_t data2, uint8_t data3, uint8_t data4) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer(data1);
  SPI.transfer(data2);
  SPI.transfer(data3);
  SPI.transfer(data4);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// TFTにコマンド+5バイトデータを送信
void ILI9342C::tftSendCommand5(uint8_t command, uint8_t data1, uint8_t data2, uint8_t data3, uint8_t data4, uint8_t data5) {
  digitalWrite(TFT_CS, LOW); // SILI9342Cセレクト
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(command);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer(data1);
  SPI.transfer(data2);
  SPI.transfer(data3);
  SPI.transfer(data4);
  SPI.transfer(data5);
  digitalWrite(TFT_CS, HIGH);  // SILI9342C解放
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
uint8_t ILI9342C::tftReadCommand(uint8_t command) {
  uint8_t data ;
  digitalWrite(TFT_CS, LOW);  //通信開始
  digitalWrite(TFT_DC, LOW); // Command mode
  // Send Command
  SPI.transfer(command);
  // Read Data
  digitalWrite(TFT_DC, HIGH); // Data mode
  data  = SPI.transfer(0);
  digitalWrite(TFT_DC, LOW); // Command mode
  digitalWrite(TFT_CS, HIGH); //通信終了

  return data ;
}

// ▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼▲▼
void ILI9342C::lcd_init() {
  // --- HARDWARE Reset
//  digitalWrite(TFT_RST, LOW);
//  delay(20);
//  digitalWrite(TFT_RST, HIGH);
//  delay(20);    // このdelayは必要

  // --- SOFTWARE Reset
  tftSendCommand(ILI9342_SOFTRESET) ;   // SOFTWARE RESET
  delay(150);                           // このdelayはなくても大丈夫かも

  tftSendCommand(ILI9342_NORON) ; //  Normal Display Mode ON
                          //    このコマンドは、表示を通常モードに戻します。
                          //    通常表示モードのオンは、部分モードのオフを意味します。
                          //    部分モードオンコマンド=(0x12)
  tftSendCommand(ILI9342_INVON) ; //  Display Inversion ON
                          //    このコマンドは、表示反転モードからリカバリーするために使用されます。
                          //    このコマンドは、フレームメモリの内容を変更しません。
                          //    このコマンドは、他の状態を変更しません。
  
  // ----- ILI9431動作モード設定 -----
  screenDir(1) ;

  // ピクセルフォーマットを1Dot/16bitに設定
  tftSendCommand1(ILI9342_COLMOD,ILI9342_COLOR16);

  // PowerOFFにならないモード
  tftSendCommand(ILI9342_SLEEP_OUT);
  delay(60);  // このdelayは必要

  // 画面表示ON
  tftSendCommand(ILI9342_DISPLAY_ON);
}

void ILI9342C::screenDir(int dir) {
  //  MY - 0x80 // Row Address Order
  //  MX - 0x40 // Column Address Order
  //  MV - 0x20 // Row / Column Exchange
  //  ML - 0x10 // Vertical Refresh Order
  //  MRGB - 0x08 // RBG - BGR
  //  MH - 0x04 // Horizontal Refresh ORDER
  uint8_t PR = 0 ;

  tftWidth = ILI9342_WIDTH ;
  tftHeight=ILI9342_HEIGHT ;
  switch(dir) {
    case 1:
      tftWidth = ILI9342_WIDTH ;
      tftHeight= ILI9342_HEIGHT ;
      PR = 0x08 ;   // MX | MRGB // DIR - 1 : 縦
      break ;
    case 3:
      tftWidth = ILI9342_WIDTH ;
      tftHeight= ILI9342_HEIGHT ;
      PR = 0xC8;    // MY | MH | MRGB // DIR - 3 : 縦
      break ;
    case 2:
      tftWidth = ILI9342_HEIGHT ;
      tftHeight= ILI9342_WIDTH ;
      PR = 0x68 ;   // MY | MX | MV | MRGB //　dir-2 : 横
      break ;
    case 4:
      tftWidth = ILI9342_HEIGHT ;
      tftHeight= ILI9342_WIDTH ;
      PR = 0xA8 ;   // MV | MRGB //　dir-4 : 横
      break ;
  }
  tftSendCommand1(ILI9342_MEMORY_ACCESS_CONTROL,PR) ;
}

// △▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽△▽
// BITマップを表示する::基本fillと同じ作り
void ILI9342C::dispBitMap(uint16_t startX , uint16_t startY , uint16_t width, uint16_t hight, uint8_t bitMap[]) {
  uint16_t endX = startX + width - 1;
  uint16_t endY = startY + hight - 1;

  digitalWrite(TFT_CS, LOW);  //通信開始

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_COLUMN_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(startX);
  SPI.transfer16(endX);

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_PAGE_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(startY);
  SPI.transfer16(endY);

  int pos = 0 ;
  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_MEMORY_WRITE);
  digitalWrite(TFT_DC, HIGH); // Command mode
  for (int row = 0 ; row < hight ; row ++) {
    // 表示データ転送
    // Arduinoだとメモリが足りなくて１画面分を一気に転送する事ができない
    // なので、1Lineずつ送信している
    for (int col = 0 ; col < width * 2 ; col ++ ) {
      SPIBuf[col] = bitMap[pos];
      pos ++ ;
    }
    SPI.transfer(SPIBuf, width * 2);
  }
  digitalWrite(TFT_DC, LOW); // Command mode
  digitalWrite(TFT_CS, HIGH); //通信終了
}

void ILI9342C::fill(uint16_t xs,uint16_t ys,uint16_t xe,uint16_t ye,uint16_t color) {
  int hight = ye - ys + 1 ;
  int width = xe - xs + 1 ;
  int num = hight * width ;
  
  digitalWrite(TFT_CS, LOW);  //通信開始

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_COLUMN_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(xs);
  SPI.transfer16(xe);

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_PAGE_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(ys);
  SPI.transfer16(ye);

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_MEMORY_WRITE);
  digitalWrite(TFT_DC, HIGH); // Command mode
  for (int i = 0 ; i < num ; i++) {
    SPI.transfer16(color);
  }

  digitalWrite(TFT_DC, LOW); // Command mode
  digitalWrite(TFT_CS, HIGH); //通信終了
}

void ILI9342C::fillScreen(uint16_t color) {
  digitalWrite(TFT_CS, LOW);  //通信開始

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_COLUMN_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(0);
  SPI.transfer16(tftWidth-1);

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_PAGE_ADDRESS_SET);
  digitalWrite(TFT_DC, HIGH); // Command mode
  SPI.transfer16(0);
  SPI.transfer16(tftHeight-1);

  digitalWrite(TFT_DC, LOW); // Command mode
  SPI.transfer(ILI9342_MEMORY_WRITE);
  digitalWrite(TFT_DC, HIGH); // Command mode
  for (uint16_t y = 0; y < tftHeight; ++y) {
    for (uint16_t x = 0; x < tftWidth; ++x) {
      SPI.transfer16(color);
    }
  }

  digitalWrite(TFT_DC, LOW); // Command mode
  digitalWrite(TFT_CS, HIGH); //通信終了
}
