#include <SD.h>
#include <SPI.h>
#include "LoadJPEG.h"
#include "Image.h"

#define SD_CS 4
#define DEBUG false

LoadJPEG::LoadJPEG( ) {
  imageBuff = NULL ;
  panelBuff = NULL ;
}

LoadJPEG::~LoadJPEG( ) {
  free(imageBuff) ;
  free(panelBuff) ;
}

void LoadJPEG::initBuff( ) {
  if (imageBuff == NULL) {
    imageBuff = (BYTE *)malloc(320*240*2) ;
#if DEBUG
    Serial.printf("malloc imageBuff %x\r\n",imageBuff);
#endif
  }
  if (panelBuff == NULL) {
    panelBuff = (BYTE *)malloc(320*240*2) ;
#if DEBUG
    Serial.printf("malloc panelBuff %x\r\n",imageBuff);
#endif
    for (int i=0;i<16;i++) {
      panelPtr[i] = &panelBuff[i*80*60*2] ;
    }
  }
  memset(imageBuff,0xFF,320*240*2) ;
  memset(panelBuff,0xFF,320*240*2) ;
}

void LoadJPEG::makePanelImage( ) {
  for (int y=0;y<4;y++) {
    for (int x=0;x<4;x++) {
      int p = y*4+x ;
      for (int v=0;v<60;v++) {
        for (int h=0;h<80;h++) {
          int dp1 = (v*80+h)*2 ;
          int dp2 = ((y*60+v)*320+x*80+h)*2 ;
          int dp3 = (y*60+v)*320+x*80+h ;
          switch(image_data_image[dp3]) {
            case 0xFF:
              panelPtr[p][dp1] = 0xFF ;
              panelPtr[p][dp1+1] = 0xFF ;
              break ;
            case 0x00:
              panelPtr[p][dp1] = 0x00 ;
              panelPtr[p][dp1+1] = 0x00 ;
              break ;
            default:
              panelPtr[p][dp1] = imageBuff[dp2] ;
              panelPtr[p][dp1+1] = imageBuff[dp2+1] ;
          }
        }
      }
    }
  }
}

void LoadJPEG::overLayImage( ) {
  int dp = 0 ;
  for (int p=0;p<320*240;p++,dp+=2) {
    switch(image_data_image[p]) {
      case 0xFF:
        imageBuff[dp] = 0xFF ;
        imageBuff[dp+1] = 0xFF ;
        break ;
      case 0x00:
        imageBuff[dp] = 0x00 ;
        imageBuff[dp+1] = 0x00 ;
        break ;
    }
  }
}

void LoadJPEG::overLayImage2( ) {
  int dp = 0 ;
  for (int p=0;p<320*240;p++,dp+=2) {
    switch(image_data_image2[p]) {
      case 0xFF:
        imageBuff[dp] = 0xFF ;
        imageBuff[dp+1] = 0xFF ;
        break ;
      case 0x00:
        imageBuff[dp] = 0x00 ;
        imageBuff[dp+1] = 0x00 ;
        break ;
    }
  }
}

UINT in_func (JDEC* jd, BYTE* buff, UINT nbyte)
{
  UINT ret = 0;
  IODEV *dev = (IODEV*)jd->device;   // Device identifier for the session (5th argument of jd_prepare function) 
  File *f2 = (File *)dev->fp;
  if (buff) {
    return (UINT)f2->read(buff, nbyte);
  } else {
    // Remove bytes from input stream 
    if(f2->seek(nbyte, SeekCur)){ //seek mode:SeekCur=1(FS.h)
      ret = nbyte;
    }else{
      ret = 0;
    }
  }
  return ret;
}
 
UINT out_func (JDEC* jd, void* bitmap, JRECT* rect)
{
  IODEV *dev = (IODEV*)jd->device;
  uint8_t *src, *dst;
  uint16_t x, y, bwd;
 
#if DEBUG
  if (rect->left == 0) {
    Serial.printf("%lu%%\r\n", (rect->top << jd->scale) * 100UL / jd->height);
  }
#endif
  src = (uint8_t*)bitmap;
  //use RGB565
  dst = dev->fbuf + 2 * (rect->top * dev->wfbuf + rect->left);;
  //bws = 2 * (rect->right - rect->left + 1);
  bwd = 2 * dev->wfbuf;
  uint8_t msb, lsb, tmp1, tmp2;
  int16_t xcnt;
  for (y = rect->top; y <= rect->bottom; y++) {
    xcnt = 0;
    for (x = rect->left; x <= rect->right; x++){
      msb = *(src++) & 0b11111000;
      tmp1 = *(src++) & 0b11111100;
      msb = msb | (tmp1 >> 5);
      tmp2 = *(src++) & 0b11111000;
      lsb = (tmp1 << 3) | (tmp2 >> 3);
      *(dst + xcnt++) = msb;
      *(dst + xcnt++) = lsb;
    }
    dst += bwd;
  }
 
  return 1;    // Continue to decompress 
}

bool LoadJPEG::LoadImage(char *fileName) {
  void *work;       /* Pointer to the decompressor work area */
  JDEC jdec;        /* Decompression object */
  JRESULT res;      /* Result code of TJpgDec API */
  IODEV devid;      /* User defined device identifier */

  pinMode(SD_CS, OUTPUT);
  if(!SD.begin(SD_CS)){
#if DEBUG
    Serial.println("Card Mount Failed");
#endif
    return false;
  }

  File file = SD.open(fileName);
  if(!file){
#if DEBUG
    Serial.println("Failed to open file for reading");
#endif
    return false;
  }
#if DEBUG
  Serial.printf("%s file size = %d\r\n", fileName, file.size());
  Serial.printf("1.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif 
  devid.fp = &file;
 
  // Allocate a work area for TJpgDec 
  work = malloc(3100);
  // Prepare to decompress 
  res = jd_prepare(&jdec, in_func, work, 3100, &devid);
  if (res == JDR_OK) {
    // Ready to dcompress. Image info is available here. 
#if DEBUG
    Serial.printf("Image dimensions: %u by %u. %u bytes used.\r\n", jdec.width, jdec.height, 3100 - jdec.sz_pool);
#endif
 
    devid.fbuf = (BYTE *)malloc(2 * jdec.width * jdec.height); //RGB565
    devid.wfbuf = jdec.width;
 
    res = jd_decomp(&jdec, out_func, 0);   // Start to decompress with 1/1 scaling 
#if DEBUG
    if (res == JDR_OK) {
      // Decompression succeeded. You have the decompressed image in the frame buffer here. 
      Serial.printf("\rOK  \n");
    } else {
      Serial.printf("Failed to decompress: rc=%d\n", res);
    }
  } else {
    Serial.printf("Failed to prepare: rc=%d\n", res);
#endif
  }
  free(work);             // Discard work area 
  work = NULL ;
 
#if DEBUG
  Serial.printf("2.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif
  delay(500);

  // --- Copy 2 ImageBuffer
  width_bmp = jdec.width;
  height_bmp = jdec.height;
  memset (imageBuff,0,320*240*2) ;

  int ps = 0 ;
  int pd = 0 ;
  for (int y=0;y<240||y<height_bmp;y++) {
    pd = y * 320 * 2;
    ps = y * width_bmp * 2 ;
    int wByte = (width_bmp<320) ? width_bmp * 2 : 320 *2 ;
    memcpy(&imageBuff[pd],&devid.fbuf[ps],wByte) ;
  }
 
  //M5.Lcd.setSwapBytes(true); //RGB<->BGR convert
  uint8_t X0 = 0, Y0 = 0;
  // M5.Lcd.pushImage((int32_t)X0, (int32_t)Y0, (uint32_t)width_bmp, (uint32_t)height_bmp, (uint16_t*)devid.fbuf);

  if(devid.fbuf){
    free(devid.fbuf);
    devid.fbuf = NULL;
#if DEBUG
    Serial.println("Free devid.fbuf OK!");
  }else{
    Serial.println("Failed free devid.fbuf.");
#endif
  }

#if DEBUG
  Serial.printf("3.Free Heap Size=%d\r\n", esp_get_free_heap_size());
#endif
  return true ;
}
