#include "TouchPanel.h"
#include "i2cRW.h"

#define I2C_ADDR_FT6336         0x38
#define ADDR_TD_STATUS          0x02
#define ADDR_TOUCH_ID1          0x05
#define ADDR_TOUCH_X1           0x03
#define ADDR_TOUCH_Y1           0x05

#define ADDR_TOUCH_ID2          0x0B
#define ADDR_TOUCH_X2           0x09
#define ADDR_TOUCH_Y2           0x0B

TouchPanel::TouchPanel( ) {
  BTN = 0 ;
  POSSTAT = STAT_RELEASE ;
  POSX = 0 ;
  POSY = 0 ;
}

void TouchPanel::resetPos( ) {
  POSSTAT = STAT_RELEASE ;
  BTN = 0 ;
}

void TouchPanel::setPos(uint16_t x,uint16_t y) {
  POSSTAT = STAT_TOUCH ;
  BTN = 0 ;
  if (y <240) {
    POSX = x / 80 ;
    POSY = y / 60 ;
  } else {
    if (30 <= x && x <= 110) { // 70±40
      BTN = 1 ; 
    }
    if (130 <= x && x <= 210) {// 170±40
      BTN = 2 ; 
    }
    if (230 <= x && x <= 310) {// 270±40
      BTN = 3 ;
    }
  }
}

bool TouchPanel::readPoint( ) {
  uint8_t readPoints = ReadByte(I2C_ADDR_FT6336,ADDR_TD_STATUS) ;
  if (readPoints == 1 || readPoints == 2) {
    if (readPoints == 1) {
        uint16_t data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X1) ;
        uint16_t data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y1) ;
        uint8_t id1 = (data2 >> 12) & 0x01 ;
        tpData[id1].stat = (tpData[id1].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id1].x = data1 & 0x0FFF ;
        tpData[id1].y = data2 & 0x0FFF ;
        tpData[~id1 & 0x01].stat = STAT_RELEASE; 
        if (tpData[id1].stat == STAT_TOUCH) {
          setPos(tpData[id1].x ,tpData[id1].y) ;
        }
    } else {
        uint16_t data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X1) ;
        uint16_t data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y1) ;
        uint8_t id1 = (data2 >> 12) & 0x01 ;
        tpData[id1].stat = (tpData[id1].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id1].x = data1 & 0x0FFF ;
        tpData[id1].y = data2 & 0x0FFF ;

        data1 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_X2) ;
        data2 = Read2Bytes(I2C_ADDR_FT6336,ADDR_TOUCH_Y2) ;
        uint8_t id2 = (data2 >> 12) & 0x01 ;
        tpData[id2].stat = (tpData[id2].stat == STAT_RELEASE) ? STAT_TOUCH : STAT_STREAM; 
        tpData[id2].x = data1 & 0x0FFF ;
        tpData[id2].y = data2 & 0x0FFF ;
        if (tpData[id2].stat == STAT_TOUCH) {
          setPos(tpData[id2].x ,tpData[id2].y) ;
        }
    }
  } else {
     tpData[0].stat = STAT_RELEASE; 
     tpData[1].stat = STAT_RELEASE; 
  }
  return (POSSTAT==STAT_TOUCH) ;
}
